from typing import *

from parsee.templates.element_schema import ElementDetectionSchema, ElementSchema
from parsee.templates.general_structuring_schema import StructuringItemSchema, GeneralQuerySchema, GeneralQueryItemSchema
from parsee.extraction.tasks.element_classification.element_model import ElementModel, AssignedElementModel
from parsee.extraction.tasks.questions.question_model import QuestionModel, AssignedQuestionModel
from parsee.extraction.models.llm_models.model_collection.chatgpt_model import ChatGPTModel
from parsee.extraction.models.llm_models.model_collection.replicate_model import ReplicateModel
from parsee.extraction.models.llm_models.model_collection.anthropic_model import AnthropicModel
from parsee.extraction.models.llm_models.model_collection.ollama_model import OllamaModel
from parsee.extraction.models.llm_models.model_collection.together_model import TogetherModel
from parsee.extraction.models.llm_models.model_collection.cohere_model import CohereModel
from parsee.extraction.models.llm_models.model_collection.mistral_model import MistralModel
from parsee.extraction.tasks.questions.question_model_llm import LLMQuestionModel
from parsee.extraction.tasks.meta_info_structuring.meta_info import MetaInfoModel
from parsee.extraction.tasks.meta_info_structuring.meta_info_llm import MetaLLMModel
from parsee.extraction.tasks.element_classification.element_model_llm import ElementModelLLM
from parsee.extraction.tasks.mappings.mapping_model import MappingModel
from parsee.extraction.tasks.mappings.mapping_model_llm import MappingModelLLM
from parsee.storage.interfaces import StorageManager
from parsee.utils.enums import ModelType
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel


def get_llm_base_model(spec: MlModelSpecification) -> LLMBaseModel:
    if spec.model_type == ModelType.GPT:
        return ChatGPTModel(spec)
    elif spec.model_type == ModelType.REPLICATE:
        return ReplicateModel(spec)
    elif spec.model_type == ModelType.ANTHROPIC:
        return AnthropicModel(spec)
    elif spec.model_type == ModelType.OLLAMA:
        return OllamaModel(spec)
    elif spec.model_type == ModelType.TOGETHER:
        return TogetherModel(spec)
    elif spec.model_type == ModelType.COHERE:
        return CohereModel(spec)
    elif spec.model_type == ModelType.MISTRAL:
        return MistralModel(spec)
    else:
        raise Exception("llm base model not found")


class ModelLoader:

    def __init__(self, storage: StorageManager):
        self.storage = storage

    def get_model_spec(self, model_id: Optional[str]) -> Union[MlModelSpecification, None]:
        filtered = [x for x in self.storage.get_available_models() if x.model_id == model_id]
        if len(filtered) == 0:
            return None
        return filtered[0]

    def get_question_model(self, model_id: Optional[str], items: List[GeneralQueryItemSchema], all_meta_items: List[StructuringItemSchema], params: Dict[str, Any]) -> Union[QuestionModel, None]:
        if model_id is None:
            return None
        elif model_id == "assigned":
            return AssignedQuestionModel(items, all_meta_items, **params)
        else:
            spec = self.get_model_spec(model_id)
            if spec is None or spec.model_type == ModelType.CUSTOM:
                # custom models have to be handled with a custom ModelLoader class
                return None
            return LLMQuestionModel(items, all_meta_items, self.storage, get_llm_base_model(self.get_model_spec(model_id)), **params)
    
    def get_element_model(self, model_id: Optional[str], items: List[ElementSchema], params: Dict[str, Any]) -> Union[ElementModel, None]:
        if model_id is None:
            return None
        elif model_id == "assigned":
            return AssignedElementModel(items, **params)
        else:
            spec = self.get_model_spec(model_id)
            if spec is None or spec.model_type == ModelType.CUSTOM:
                # custom models have to be handled with a custom ModelLoader class
                return None
            return ElementModelLLM(items, self.storage, get_llm_base_model(spec), **params)
    
    def get_meta_model(self, model_id: Optional[str], items: List[StructuringItemSchema], params: Dict[str, Any]) -> Union[MetaInfoModel, None]:
        if model_id is None:
            return None
        elif model_id == "assigned":
            # for assigned, no model is needed
            return None
        else:
            spec = self.get_model_spec(model_id)
            if spec is None or spec.model_type == ModelType.CUSTOM:
                # custom models have to be handled with a custom ModelLoader class
                return None
            return MetaLLMModel(items, get_llm_base_model(spec), self.storage, **params)
    
    def get_mapping_model(self, model_id: Optional[str], items: List[ElementSchema], params: Dict[str, Any]) -> Union[MappingModel, None]:
        if model_id is None:
            return None
        elif model_id == "assigned":
            return None # TODO
        else:
            spec = self.get_model_spec(model_id)
            if spec is None or spec.model_type == ModelType.CUSTOM:
                # custom models have to be handled with a custom ModelLoader class
                return None
            return MappingModelLLM(items, self.storage, get_llm_base_model(spec), **params)


def question_models_from_schema(schema: GeneralQuerySchema, all_meta_items: List[StructuringItemSchema], model_loader: ModelLoader, params: Dict[str, Any]) -> List[QuestionModel]:

    # group items by model
    items_by_model: Dict[str, List[GeneralQueryItemSchema]] = {}
    for item in schema.items:
        if item.model not in items_by_model:
            items_by_model[item.model] = []
        items_by_model[item.model].append(item)

    models: List[QuestionModel] = []
    for model_name, items in items_by_model.items():
        model = model_loader.get_question_model(model_name, items, all_meta_items, params)
        if model is None and model_name != "assigned":
            raise Exception(f"following model was not found: {model_name}")
        if model is not None:
            models.append(model)
    return models


def element_models_from_schema(schema: ElementDetectionSchema, model_loader: ModelLoader, params: Dict[str, any]) -> List[ElementModel]:

    # group items by model
    items_by_model: Dict[str, List[ElementSchema]] = {}
    for item in schema.items:
        if item.model not in items_by_model:
            items_by_model[item.model] = []
        items_by_model[item.model].append(item)

    models: List[ElementModel] = []
    for model_name, items in items_by_model.items():
        model = model_loader.get_element_model(model_name, items, params)
        if model is None and model_name != "assigned":
            raise Exception(f"following model was not found: {model_name}")
        if model is not None:
            models.append(model)
    return models


def meta_models_from_items(meta_items: List[StructuringItemSchema], model_loader: ModelLoader, params: Dict[str, str]) -> List[MetaInfoModel]:

    # group items by model
    items_by_model: Dict[str, List[StructuringItemSchema]] = {}
    for item in meta_items:
        if item.model not in items_by_model:
            items_by_model[item.model] = []
        items_by_model[item.model].append(item)

    models: List[MetaInfoModel] = []
    for model_name, items in items_by_model.items():
        model = model_loader.get_meta_model(model_name, items, params)
        if model is None and model_name != "assigned":
            raise Exception(f"following model was not found: {model_name}")
        if model is not None:
            models.append(model)
    return models


def mapping_models_from_schema(schema: ElementDetectionSchema, model_loader: ModelLoader, params: Dict[str, str]) -> List[MappingModel]:

    # group items by model
    items_by_model: Dict[str, List[ElementSchema]] = {}
    for item in schema.items:
        if item.mapRows is not None:
            if item.mappingModel not in items_by_model:
                items_by_model[item.mappingModel] = []
            items_by_model[item.mappingModel].append(item)

    models: List[MappingModel] = []
    for model_name, items in items_by_model.items():
        model = model_loader.get_mapping_model(model_name, items, params)
        if model is None and model_name != "assigned":
            raise Exception(f"following model was not found: {model_name}")
        if model is not None:
            models.append(model)
    return models



from typing import *

from parsee.extraction.extractor_elements import StandardDocumentFormat
from parsee.extraction.tasks.questions.question_model import QuestionModel
from parsee.extraction.tasks.questions.features import GeneralQueriesPromptBuilder, MAIN_QUESTION_STR
from parsee.extraction.extractor_dataclasses import ParseeAnswer, ParseeMeta
from parsee.storage.interfaces import StorageManager
from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel
from parsee.templates.general_structuring_schema import StructuringItemSchema, GeneralQueryItemSchema
from parsee.extraction.models.llm_models.structuring_schema import get_prompt_schema_item
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.utils.helper import parse_json_dict


class LLMQuestionModel(QuestionModel):

    def __init__(self, items: List[GeneralQueryItemSchema], meta_items: List[StructuringItemSchema], storage: StorageManager, llm: LLMBaseModel, **kwargs):
        super().__init__(items, meta_items)
        self.storage = storage
        self.llm = llm
        self.model_name = llm.spec.model_id
        self.prompt_builder = GeneralQueriesPromptBuilder(storage)

    def parse_prompt_answer(self, item: GeneralQueryItemSchema, prompt_answer: str, total_elements: Optional[int], document: Optional[StandardDocumentFormat]) -> List[ParseeAnswer]:

        json_data = parse_json_dict(prompt_answer)

        if json_data is None:
            return []

        answer_blocks = []
        if "answers" in json_data and isinstance(json_data["answers"], list):
            answer_blocks = json_data["answers"]
        elif MAIN_QUESTION_STR in json_data:
            answer_blocks = [json_data]

        sources = []
        if "sources" in json_data and isinstance(json_data["sources"], list):
            sources = [int(x) for x in json_data["sources"] if str(x).isdigit() and (total_elements is None or 0 <= int(x) <= total_elements)]

        sources_full = [document.elements[el_idx].source for el_idx in sources] if document is not None else []

        output = []
        for main_and_meta in answer_blocks:

            if not isinstance(main_and_meta, dict):
                continue

            main_answer = None
            parse_successful = False
            detected_meta = []

            for key, val in main_and_meta.items():

                if key is None:
                    continue

                val = str(val)

                if MAIN_QUESTION_STR in key:
                    prompt_item = get_prompt_schema_item(item)
                    main_answer, parse_successful = prompt_item.get_value(val)
                else:
                    meta_items_filtered = [x for x in self.meta if x.id == key]
                    if len(meta_items_filtered) > 0:
                        meta_item = meta_items_filtered[0]
                        prompt_item = get_prompt_schema_item(meta_item)
                        parsed, parse_successful = prompt_item.get_value(val)

                        detected_meta.append(ParseeMeta(self.model_name, 0, sources_full, meta_item.id, parsed, 0.8 if parse_successful else 0))

            if main_answer is not None:

                output.append(ParseeAnswer(self.model_name, sources_full, item.id, main_answer, prompt_answer, parse_successful, detected_meta))

        return output

    def predict_for_prompt(self, prompt: Prompt, schema_item: GeneralQueryItemSchema, max_element_index: Optional[int], document: Optional[StandardDocumentFormat]) -> List[ParseeAnswer]:
        prompt_answer, amount = self.llm.make_prompt_request(prompt)
        self.storage.log_expense(self.llm.spec.model_id, amount, schema_item.id)
        return self.parse_prompt_answer(schema_item, prompt_answer, max_element_index, document)

    def predict_answers(self, document: StandardDocumentFormat) -> List[ParseeAnswer]:

        answers: List[ParseeAnswer] = []
        for schema_item in self.items:
            relevant_elements = self.prompt_builder.get_relevant_elements(schema_item, document)
            prompt = self.prompt_builder.build_prompt(schema_item, self.meta, document, relevant_elements, self.llm.spec.multimodal, self.llm.spec.max_images, self.llm.spec.max_image_pixels)
            answers += self.predict_for_prompt(prompt, schema_item, len(document.elements), document)

        return answers



from typing import *

from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl, FileReference


class VectorStore:

    def find_closest_elements(self, document: StandardDocumentFormat, search_element_title: str, keywords: str, tables_only: bool = True) -> List[ExtractedEl]:
        raise NotImplementedError

    def sort_identifiers_by_relevance(self, source_identifiers: Set[str], search_query: str) -> Set[str]:
        raise NotImplementedError


from typing import *
from decimal import Decimal

from parsee.extraction.extractor_dataclasses import ParseeBucket
from parsee.templates.mappings import MappingSchema


def calc_buckets(key_value_pairs: List[Tuple[str, any]], mappings: List[ParseeBucket], mapping_schema: MappingSchema) -> Dict[str, Decimal]:

    if len(mappings) != len(key_value_pairs):
        return {}

    output = {}

    # TODO: improve

    for k, mapping in enumerate(mappings):

        if mapping is None:
            continue

        if mapping.bucket_id not in output:
            output[mapping.bucket_id] = Decimal(0)

        val = key_value_pairs[k][1]
        if val is not None:
            output[mapping.bucket_id] += Decimal(val)

    return output


def get_table_signature(key_value_pairs: List[Tuple[str, any]]) -> str:

    return ", ".join([f"{k}: {x[0]}" for k, x in enumerate(key_value_pairs)])



from typing import *

from parsee.datasets.dataset_dataclasses import DatasetRow, DatasetColumn, Transformation, ColumnData, ColumnSettings
from parsee.extraction.models.model_dataclasses import MlModelSpecification


class DatasetReader:

    preprocessing_fun: Optional[Callable] = None
    indices: List[int]
    total_rows: int

    def __init__(self, indices: List[int]):
        self.reassign_indices(indices)

    def reassign_indices(self, indices: List[int]):
        self.indices = list(sorted(indices))
        self.total_rows = len(self.indices)

    def row_generator(self) -> Generator[Tuple[DatasetRow, int], None, None]:
        raise NotImplementedError

    def infinite_final_data_generator(self, column_data: List[ColumnData], batch_size: int) -> Generator[any, None, None]:
        raise NotImplementedError

    def apply_skip_condition_fun(self, fun: Callable, args: Dict):
        raise NotImplementedError

    def set_preprocessing_fun(self, fun: Callable):
        self.preprocessing_fun = fun

    def get_columns(self) -> List[DatasetColumn]:
        raise NotImplementedError

    def stream_column(self, col_name: str) -> Generator[any, None, None]:
        raise NotImplementedError

    def undersample(self, fun: Callable):
        raise NotImplementedError


class ModelReader:

    def load_model(self, model_specification: MlModelSpecification) -> Tuple[any, Dict[str, ColumnSettings]]:
        raise NotImplementedError